use std::borrow::Cow;

/// Removes all whitespace from a string.
///
/// Example: "foo bar" => "foobar" and "foo \n bar" => "foobar".
#[allow(dead_code)]
pub(crate) fn coalesce_whitespace(input: &str) -> Cow<'_, str> {
    let mut state = CoalesceWhitespace::Normal;
    for (offset, c) in input.char_indices() {
        match (&mut state, c) {
            (CoalesceWhitespace::Normal, ' ' | '\t' | '\r' | '\n') => {
                let mut ret = String::with_capacity(input.len());
                ret.push_str(&input[..offset]);
                ret.push(' ');
                state = CoalesceWhitespace::HasWhitespace {
                    in_whitespace: true,
                    ret,
                };
            }
            (CoalesceWhitespace::Normal, _) => {}
            (
                CoalesceWhitespace::HasWhitespace { in_whitespace, ret },
                ' ' | '\t' | '\r' | '\n',
            ) => {
                if !*in_whitespace {
                    *in_whitespace = true;
                    ret.push(' ');
                }
            }
            (CoalesceWhitespace::HasWhitespace { in_whitespace, ret }, _) => {
                *in_whitespace = false;
                ret.push(c);
            }
        }
    }
    match state {
        CoalesceWhitespace::Normal => Cow::Borrowed(input),
        CoalesceWhitespace::HasWhitespace {
            in_whitespace: _,
            ret,
        } => Cow::Owned(ret),
    }
}

enum CoalesceWhitespace {
    Normal,
    HasWhitespace { in_whitespace: bool, ret: String },
}

pub(crate) struct FullLinesIterator<'i> {
    inner: Option<&'i str>,
}

impl<'i> Iterator for FullLinesIterator<'i> {
    type Item = &'i str;

    fn next(&mut self) -> Option<Self::Item> {
        if let Some(inner) = self.inner {
            for (i, c) in inner.char_indices() {
                if c == '\n' {
                    let (current_line, remainder) = inner.split_at(i + 1);
                    self.inner = Some(remainder);
                    return Some(current_line);
                }
            }
            if inner.is_empty() {
                None
            } else {
                let final_unterminated_line = Some(inner);
                self.inner = None;
                final_unterminated_line
            }
        } else {
            None
        }
    }

    fn size_hint(&self) -> (usize, Option<usize>) {
        if let Some(inner) = self.inner {
            let line_count = {
                inner.bytes().filter(|b| *b == b'\n').count()
                    + (if inner.ends_with('\n') { 0 } else { 1 })
            };
            (line_count, Some(line_count))
        } else {
            (0, Some(0))
        }
    }
}

impl ExactSizeIterator for FullLinesIterator<'_> {}

pub(crate) trait GetFullLines {
    fn full_lines(&self) -> FullLinesIterator<'_>;
}

impl<'s, S: AsRef<str>> GetFullLines for S {
    fn full_lines(&self) -> FullLinesIterator<'_> {
        let inner = self.as_ref();
        FullLinesIterator {
            inner: if inner.is_empty() { None } else { Some(inner) },
        }
    }
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn test_full_lines_iterator() {
        {
            let input = "foo\nbar\nbaz";
            assert_eq!(input.lines().count(), input.full_lines().count());
            assert_eq!(input.lines().count(), input.full_lines().len());
            assert_eq!(input.lines().collect::<Vec<_>>(), vec!["foo", "bar", "baz"]);
            assert_eq!(
                input.full_lines().collect::<Vec<_>>(),
                vec!["foo\n", "bar\n", "baz"]
            );
        }

        {
            // Trailing newline
            let input = "foo\nbar\nbaz\n";
            assert_eq!(input.lines().count(), input.full_lines().count());
            assert_eq!(input.lines().count(), input.full_lines().len());
            assert_eq!(input.lines().collect::<Vec<_>>(), vec!["foo", "bar", "baz"]);
            assert_eq!(
                input.full_lines().collect::<Vec<_>>(),
                vec!["foo\n", "bar\n", "baz\n"]
            );
        }

        {
            // Leading newline
            let input = "\nfoo\nbar\nbaz";
            assert_eq!(input.lines().count(), input.full_lines().count());
            assert_eq!(input.lines().count(), input.full_lines().len());
            assert_eq!(
                input.lines().collect::<Vec<_>>(),
                vec!["", "foo", "bar", "baz"]
            );
            assert_eq!(
                input.full_lines().collect::<Vec<_>>(),
                vec!["\n", "foo\n", "bar\n", "baz"]
            );
        }

        {
            // Double newline
            let input = "foo\nbar\n\nbaz";
            assert_eq!(input.lines().count(), input.full_lines().count());
            assert_eq!(input.lines().count(), input.full_lines().len());
            assert_eq!(
                input.lines().collect::<Vec<_>>(),
                vec!["foo", "bar", "", "baz"]
            );
            assert_eq!(
                input.full_lines().collect::<Vec<_>>(),
                vec!["foo\n", "bar\n", "\n", "baz"]
            );
        }

        {
            // Empty
            let input = "";
            assert_eq!(input.lines().count(), input.full_lines().count());
            assert_eq!(input.lines().count(), input.full_lines().len());
            assert_eq!(input.lines().collect::<Vec<_>>(), Vec::<&str>::new());
            assert_eq!(input.full_lines().collect::<Vec<_>>(), Vec::<&str>::new());
        }
    }
}
